// Code based on Sivan Toledo's original published at http://www.tau.ac.il/~stoledo/lego/AVR-ir/
// Modified by Jonathan Oakley

//typedef unsigned int uint16_t;
//typedef unsigned char uint8_t;

#ifndef F_CPU
//#define F_CPU		 8000000UL
//#define F_CPU	 	11059200UL
// #define F_CPU	14745600UL
#endif

#ifndef UART_BAUD_RATE
#define UART_BAUD_RATE 38400UL
//#define UART_BAUD_RATE 1200UL
//#define UART_BAUD_RATE 9600UL
#endif

#if F_CPU<14000000UL
#define DEBUG_LED_EN  {DDRD |= 0x80;}
#define DEBUG_LED_ON  {PORTD &= ~0x80;}
#define DEBUG_LED_OFF {PORTD |= 0x80;}
#else
#define DEBUG_LED_EN  {DDRB |= 0x08;}
#define DEBUG_LED_ON  {PORTB |=  0x08;}
#define DEBUG_LED_OFF {PORTB &= ~0x08;}

#endif

/*
#define T0_PRESCALE 8
#define T1_PRESCALE 64
#define TX_PERIOD 48
#define TX_HIGH 20
#define SND_INIT_HIGH 	223
#define SND_LOW 		60
#define SND_LONG 		110
#define SND_SHORT 		40
#define SND_QUIET 		15000U
#define RCV_PERIOD 		768
#define RCV_LED_TIMEOUT 20000UL
#define RCV_PULSE_MIN 	50
#define RCV_PULSE_MAX 	80
#define RCV_START_MIN 	190
#define RCV_START_MAX 	235
#define RCV_BIT_MIN 	60
#define RCV_BIT_MID 	143
#define RCV_BIT_MAX 	200
*/

/*
 * Above are Sivan Toledo's original settings for a
 * 14.7456MHz clock
 * 
 * According to Lego documentation, we actually have
 * Carrier of 38kHz
 * Pulse of 6 cycles
 * Pulse interval of 16 cycles => low
 * Pulse interval of 27 cycles => high
 * Pulse interval of 45 cycles => start or stop
 * 
 */
#define T0_PRESCALE 8
#define TX_PERIOD (F_CPU/T0_PRESCALE/38000UL)
#define TX_HIGH (F_CPU/T0_PRESCALE*1/5/38000UL)

// Lego definitions in 38kHz cycles...
#define N_PULSE		6
#define NX_LOW 		10
#define NX_HIGH 	21
#define NX_START	39
// ... and receive time thresholds, in microseconds
#define T_MIN		316
#define T_LH		526
#define T_HS		947
#define T_MAX		1579


#define T1_PRESCALE 64
#define SND_INIT_HIGH 	(F_CPU/T1_PRESCALE*NX_START/38000UL)
#define SND_LOW 		(F_CPU/T1_PRESCALE*N_PULSE/38000UL)
#define SND_LONG 		(F_CPU/T1_PRESCALE*NX_HIGH/38000UL)
#define SND_SHORT 		(F_CPU/T1_PRESCALE*NX_LOW/38000UL)
#define SND_QUIET 		(F_CPU/T1_PRESCALE*3333/1000000UL)
#define RCV_PERIOD 		(F_CPU/T1_PRESCALE*1667/1000000UL)
#define RCV_LED_TIMEOUT 20000UL

#define SND_NO_BURST	(F_CPU/T1_PRESCALE*250/1000)

// Pulse lengths look longer than expected, allow 3-10 cycles
#define RCV_PULSE_MIN 	(F_CPU/T1_PRESCALE*75/1000000UL)
#define RCV_PULSE_MAX 	(F_CPU/T1_PRESCALE*275/1000000UL)

#define RCV_START_MIN 	(F_CPU/T1_PRESCALE*T_HS/1000000UL)
#define RCV_START_MAX 	(F_CPU/T1_PRESCALE*T_MAX/1000000UL)
#define RCV_BIT_MIN 	(F_CPU/T1_PRESCALE*T_MIN/1000000UL)
#define RCV_BIT_MID 	(F_CPU/T1_PRESCALE*T_LH/1000000UL)
#define RCV_BIT_MAX 	(F_CPU/T1_PRESCALE*T_HS/1000000UL)

// IR state machine states - order IS important
typedef enum {
// in receive, but timeouts indicate duff data
	rcv_initial_low=1,
	rcv_initial_high,
	rcv_second_low,
	rcv_low,
	rcv_quiet,
// in receive, timeout may mean packet complete
	rcv_high,
// waiting for new received IR data
	rcv_done,
	rcv_idle,

// waiting for next re-transmit of Tx data
	snd_quiet,
	snd_idle,
	snd_initial_low,
	snd_initial_high,
	snd_second_low,
	snd_low,
	snd_high,
	snd_done
} PFstate;

typedef struct {
	uint8_t  state;
	uint16_t newrx;
} PFstateS;

typedef struct {
	uint16_t msg;
	uint8_t cnt;
} PFmsgS;

typedef union {
	uint8_t u8[3];
	PFmsgS msg;
	PFstateS PFstate;
} PFmsg;

#define PF_MSG_SLOTS 16

// Packet shapes as defined by Lego:
#define PF_INVALID 0x0000

typedef struct {
	uint8_t lrc:4;
	uint8_t nibble3:4;
	uint8_t nibble2:4;
	uint8_t nibble1:4;
} PFlrc;

typedef enum {
	pf_float_d=0, 
	pf_forward,
	pf_reverse,
	pf_brake_d
} PFOutputMode;

typedef struct {
	uint8_t lrc:4;

	PFOutputMode aout:2;
	PFOutputMode bout:2;

	uint8_t mode:3;
	uint8_t address:1;
	uint8_t channel:2;
	uint8_t escape:1;
	uint8_t toggle:1;
} PFdirect;

typedef enum {
	pf_A=0, 
	pf_B
} PFOutput;

typedef enum {
	pf_C1=0, 
	pf_C2
} PFPin;

typedef enum {
	pf_no_change=0, 
	pf_clear,
	pf_set,
	pf_toggle
} PFFunction;

typedef enum {
	pf_ignore=0,
	pf_direct=1,
	pf_continuous,
	pf_timeout
} PFMode;

typedef struct {
	uint8_t lrc:4;

	PFFunction	function:2;
	PFPin 		pin:1;
	PFOutput	output:1;

	PFMode  mode:3;
	uint8_t address:1;
	uint8_t channel:2;
	uint8_t escape:1;
	uint8_t toggle:1;
} PFcontinuous;

typedef struct {
	uint8_t lrc:4;

	PFFunction	function:2;
	PFPin 		pin:1;
	PFOutput	output:1;

	PFMode  mode:3;
	uint8_t address:1;
	uint8_t channel:2;
	uint8_t escape:1;
	uint8_t toggle:1;
} PFtimeout;

typedef enum {
	pf_PWM=0, 
	pf_CST		// Clear/Set/Toggle
} PFSingleMode;

typedef enum {
	pf_float=0,
	pf_fwd1,
	pf_fwd2,
	pf_fwd3,
	pf_fwd4,
	pf_fwd5,
	pf_fwd6,
	pf_fwd7,
	pf_brake,
	pf_rev7,
	pf_rev6,
	pf_rev5,
	pf_rev4,
	pf_rev3,
	pf_rev2,
	pf_rev1
} PFPWMData;

typedef enum {
	pf_cC1cC2, // set and clear C1 and C2
	pf_sC1cC2,
	pf_cC1sC2,
	pf_sC1sC2,
	pf_incPWM, // increment/decrement PWM
	pf_decPWM,
	pf_fwdTO,  // full fwd/rev with timeout
	pf_revTO,
	pf_tglTO   // full fwd/rev toggle
} PFCSTData;

typedef union {
	PFPWMData PWM;
	PFCSTData CST;
} PFDataU;


typedef struct {
	uint8_t lrc:4;

	PFPWMData	 data:4;
	PFOutput	 output:1;
	PFSingleMode mode:1;

	uint8_t one:1;
	uint8_t address:1;
	uint8_t channel:2;
	uint8_t escape:1;
	uint8_t toggle:1;
} PFsinglePWM;

typedef struct {
	uint8_t lrc:4;

	PFCSTData	 data:4;
	PFOutput	 output:1;
	PFSingleMode mode:1;

	uint8_t one:1;
	uint8_t address:1;
	uint8_t channel:2;
	uint8_t escape:1;
	uint8_t toggle:1;
} PFsingleCST;


typedef struct {
	uint8_t lrc:4;

	PFPWMData	PWMa:4;
	PFPWMData	PWMb:4;
	PFOutput	 output:1;
	PFSingleMode mode:1;

	uint8_t channel:2;
	uint8_t escape:1;
	uint8_t address:1;
} PFPWM;

typedef union {
	uint16_t	 u16;
	PFlrc    	 lrc;
	PFdirect 	 direct;
	PFcontinuous continuous;
	PFtimeout	 timeout;
	PFsinglePWM	 singlePWM;
	PFsingleCST	 singleCST;
	PFPWM   	 pwm;
} PFcmdU;


extern volatile uint32_t idleperiod;

// extern volatile int8_t  pf_repeats;
extern volatile uint16_t pf_cmd;
extern volatile uint16_t pf_cmd2;
extern volatile uint16_t pf_cmd_recv,pf_newrx;
extern volatile uint8_t pf_cmd_cnt;
extern volatile uint16_t pf_cmd_recv_valid;
//extern volatile uint8_t  nperiods;
// extern volatile uint16_t periods[];
extern volatile PFstate state;
extern volatile uint32_t long_timer;
extern volatile uint8_t q_rpt;

extern volatile uint8_t i2c_processed;
extern volatile PFmsg i2c_out_msgA[],i2c_snd_msgA[];


extern void pf_send_start();
extern void pf_send(uint16_t,uint8_t);
extern void pf_init();
extern void pf_receive();

extern void i2c_pullup_config();
extern void i2c_slave_init() ;
