// Code based on Sivan Toledo's original published at http://www.tau.ac.il/~stoledo/lego/AVR-ir/
// Modified by Jonathan Oakley

#include <stdlib.h>
#include <stdint.h>

#include <avr/io.h>
#include <avr/interrupt.h>
//#include <avr/signal.h>
#include <avr/pgmspace.h>
#include "uart.h"
#include "PFir.h"

volatile uint16_t pf_cmd2;

volatile uint8_t pf_repeats = 0,q_rpt;

volatile uint16_t pf_cmd,q_cmd,pf_newrx=0;
static volatile uint16_t pf_mask;
static volatile uint16_t pf_cmd_prev;
static volatile uint16_t pf_cmd_prev;
volatile uint16_t pf_cmd_recv;
volatile uint8_t pf_cmd_cnt=0,pf_chan;
volatile uint16_t pf_cmd_recv_valid = 0;
static volatile uint16_t period;
volatile uint32_t long_timer=SND_NO_BURST;

//volatile uint8_t  nperiods = 0;
//volatile uint16_t periods[64];

volatile PFstate state, idle_state=rcv_idle;
PFcmdU pfcmd;


//************ ALL THIS NEEDS TO BE FCPU-DEPENDENT! ***************
ISR(TIMER1_COMPA_vect)
{
#if 1
	// uint16_t xxx = TCNT1;
	
	TCNT1 -= OCR1A;
	long_timer+=OCR1A;

//				DEBUG_LED_OFF; // debug LED off

	if (state<rcv_high) // unexpected timeout in receive mode
	{
		pf_mask=1;		// flag duff data
		state=rcv_high; // and exit receive
	}

	switch (state)
	{
		case snd_idle:
			break;
			
		case snd_initial_low:
//			nperiods = 0;
//			periods[ nperiods++ ] = xxx;
			TCCR0A &= ~_BV(COM0B1);
			PORTD &= ~0x20;
			OCR1A = SND_INIT_HIGH; // initial high period
			state = snd_initial_high;
			break;
			
		case snd_initial_high:
//			periods[ nperiods++ ] = xxx;
			TCCR0A |= _BV(COM0B1);
			OCR1A = SND_LOW; // low period
			state = snd_second_low;
			break;
			
		case snd_second_low:
		case snd_low:
//			periods[ nperiods++ ] = xxx;
			// no signal
			TCCR0A &= ~_BV(COM0B1);
			PORTD &= ~0x20;
			if (pf_cmd & 0x8000)
				OCR1A = SND_LONG; // long period
			else
				OCR1A = SND_SHORT; // long period
			if (pf_mask)
				state = snd_high;
			else
			{
				idle_state = state = snd_quiet; // stop sending, but we may repeat
				OCR1A = SND_QUIET;
			}
			pf_mask <<= 1;
			pf_cmd	<<= 1;
			break;
			
		case snd_high:
//			periods[ nperiods++ ] = xxx;
			TCCR0A |= _BV(COM0B1);
			OCR1A = SND_LOW; // low period
			state = snd_low;
			break;
			
		case snd_quiet:
			if (pf_repeats) // end of quiet, but we may want to repeat
			{
				if (pf_repeats != UINT8_MAX)
					pf_repeats--;
			}

			if (pf_repeats) // still a repeat left to do, so do it!
			{
				pf_cmd = pf_cmd_prev;
				pf_send_start();
			}
			else	// repeats all done
			{
				idle_state = state = rcv_idle;
				idleperiod = 0;
				OCR1A = RCV_PERIOD; // set for receive operations
			}
			break;
			
		case snd_done:
			break;

		// intermediate "Rx quiet" state - if we timeout we can go back to
		// idling or do a re-transmit, but we might just still be in a 
		// corrupt packet, so give things a chance to settle down
		case rcv_done:
			idleperiod=0;
			state=idle_state;
			break;
			
		case rcv_idle:
			if (q_rpt)
			{
				pf_repeats=q_rpt;
				pf_cmd=q_cmd;
				pf_send_start();
				q_rpt=0;
			}

			idleperiod += RCV_PERIOD;
			if (idleperiod>RCV_LED_TIMEOUT) // idle a while, LED off
				DEBUG_LED_OFF
			break;
			
		case rcv_high: // timed out waiting for a pulse, may be end of message
			if (!pf_mask) // if it is a valid message, store it
			{
				pf_cmd_recv=pf_cmd;		// buffer received data for foreground task
				pf_cmd_recv_valid = 1; 	// and flag that it exists
				state=idle_state; 		// done OK, back to what we were doing before
			}
			else
				state = rcv_done; // was done, but not OK
			break;
			
		default:
			break;
	}
#endif
}

ISR(TIMER1_CAPT_vect)
{
		DEBUG_LED_ON; //debug LED on
	long_timer+=ICR1;

	switch (state)
	{
		case rcv_done:
		case snd_quiet:
			OCR1A=RCV_PERIOD;	// set up for receiving interleaved with sending
		case rcv_idle:
			idleperiod += ICR1;
			TCCR1B |= _BV(ICES1); // wait for rising edge
			state = rcv_initial_low;
			pf_mask=1;			// flag that timeout doesn't represent complete packet
			TCNT1 = TCNT1 - ICR1;
			break;
			
		case rcv_initial_low:
			period = ICR1;
			if (period >= RCV_PULSE_MIN && period <= RCV_PULSE_MAX)
				state = rcv_initial_high;
			else
				state = rcv_done;
			TCCR1B &= ~(_BV(ICES1)); // wait for falling edge
			TCNT1 = TCNT1 - ICR1;
			break;
			
		case rcv_initial_high:
			period = ICR1;
			if (period >= RCV_START_MIN && period <= RCV_START_MAX)
			{
				state = rcv_second_low;
				pf_mask=16;		// expect 16 more falling edges
			}
			else
				state = rcv_initial_low; // try again ...
			TCCR1B |= _BV(ICES1); // wait for rising edge
			TCNT1 = TCNT1 - ICR1;
			break;
			
		case rcv_second_low:
			period = ICR1;
			if (period >= RCV_PULSE_MIN && period <= RCV_PULSE_MAX)
			{
				pf_cmd = 0;
				state = rcv_high;
			}
			else
				state = rcv_done;
			TCCR1B &= ~(_BV(ICES1)); // wait for falling edge
			TCNT1 = TCNT1 - ICR1;
			break;
			
		case rcv_low:
			period += ICR1;
			pf_cmd <<= 1;

			if (period >= RCV_BIT_MIN && period <= RCV_BIT_MID)
				state = rcv_high;
			else
				if (period > RCV_BIT_MID && period <= RCV_BIT_MAX)
				{
					pf_cmd |= 0x0001;
					state = rcv_high;
				}
				else
					state = rcv_done;
			TCCR1B &= ~(_BV(ICES1)); // wait for falling edge
			TCNT1 = TCNT1 - ICR1;
			break;
			
		case rcv_high:
			period = ICR1;
			pf_mask--;		// count incoming pulses
			TCCR1B |= _BV(ICES1); // wait for rising edge
			state = rcv_low;
			TCNT1 = TCNT1 - ICR1;
			break;
	
//		case rcv_done:
//			idleperiod=0;
//			state=idle_state; // back to receive idle or waiting to re-transmit
//			break;
			
		default:
			DEBUG_LED_OFF; //debug LED off - can't be receiving now!
			break;
	}

}

void pf_init()
{
	cli();

/*
 *	Set timer 0 to generate the 38 KHz square wave
 */

	TCCR0A = 0 //	_BV(COM0B1) 	// set at bottom, clear on match
					 | _BV(WGM00)	// two bits of fast PWM
					 | _BV(WGM01);

	TCCR0B =	 _BV(WGM02) 		// finish waveform generation
					 | _BV(CS01); 	// divide by 8; gives about 49 counts per 38KHz cycle

	OCR0A = TX_PERIOD; 	// period
	OCR0B = TX_HIGH; 	// high time within period

	DDRD |= 0x20;	 // set pin PD5/OC0B to output
	PORTD &= ~0x20; 	// set to zero for now.

	DEBUG_LED_EN; // debug LED drive
	DEBUG_LED_OFF; // active low

	// from receive init

	TCCR1B =	_BV(ICNC1) 					// noise canceler
											// not setting ICES1, so falling edge
				 | _BV(CS11) | _BV(CS10) 	// 011, meaning divide by 64
				 ;

	TIMSK1 = _BV(ICIE1)	 		// enable capture interrupt
				 | _BV(TOIE1)	// enable overflow interrupt
				 | _BV(OCIE1A); // enable compare-match-A interrupt


	TCNT1 = 0;

	OCR1A = RCV_PERIOD; // maximum high before we declare detection

	// from send-init

/*
 * set timer 1 to generate the pulse-distance modulation
 */

	state = rcv_idle;

	sei();
}



void pf_send_start()
{
	//TIMSK1 = _BV(ICIE1)	// enable capture interrupt
	//			 | _BV(TOIE1); // enable overflow interrupt

	state = snd_initial_low;
	pf_mask = 0xff;
	pf_cmd_prev = pf_cmd;

	OCR1A = SND_LOW;
	TCNT1 = 0;

	TCCR0A |= _BV(COM0B1); // set at bottom, clear on match

//	TIMSK1 = _BV(OCIE1A); // enable compare-match-A interrupt
}

void pf_send(uint16_t cmd,uint8_t repeats)
{
	q_cmd=cmd;
	q_rpt=repeats;
}
